from builtins import range
from future.utils import viewitems, viewvalues

from miasm.expression.expression import *
from miasm.ir.ir import Lifter, IRBlock, AssignBlock
from miasm.arch.arm.arch import mn_arm, mn_armt
from miasm.arch.arm.regs import *

from miasm.jitter.csts import EXCEPT_DIV_BY_ZERO, EXCEPT_INT_XX

coproc_reg_dict = {
        ("p15", "c0", 0, "c0", 0): MIDR,
        ("p15", "c0", 0, "c0", 1): CTR,
        ("p15", "c0", 0, "c0", 2): TCMTR,
        ("p15", "c0", 0, "c0", 3): TLBTR,
        ("p15", "c0", 0, "c0", 4): MIDR,
        ("p15", "c0", 0, "c0", 5): MPIDR,
        ("p15", "c0", 0, "c0", 6): REVIDR,
        ("p15", "c0", 0, "c0", 7): MIDR,

        ("p15", "c0", 0, "c1", 0): ID_PFR0,
        ("p15", "c0", 0, "c1", 1): ID_PFR1,
        ("p15", "c0", 0, "c1", 2): ID_DFR0,
        ("p15", "c0", 0, "c1", 3): ID_AFR0,
        ("p15", "c0", 0, "c1", 4): ID_MMFR0,
        ("p15", "c0", 0, "c1", 5): ID_MMFR1,
        ("p15", "c0", 0, "c1", 6): ID_MMFR2,
        ("p15", "c0", 0, "c1", 7): ID_MMFR3,

        ("p15", "c0", 0, "c2", 0): ID_ISAR0,
        ("p15", "c0", 0, "c2", 1): ID_ISAR1,
        ("p15", "c0", 0, "c2", 2): ID_ISAR2,
        ("p15", "c0", 0, "c2", 3): ID_ISAR3,
        ("p15", "c0", 0, "c2", 4): ID_ISAR4,
        ("p15", "c0", 0, "c2", 5): ID_ISAR5,

        ("p15", "c0", 1, "c0", 0): CCSIDR,
        ("p15", "c0", 1, "c0", 1): CLIDR,
        ("p15", "c0", 1, "c0", 7): AIDR,

        ("p15", "c0", 2, "c0", 0): CSSELR,

        ("p15", "c0", 4, "c0", 0): VPIDR,
        ("p15", "c0", 4, "c0", 5): VMPIDR,

        ("p15", "c1", 0, "c0", 0): SCTLR,
        ("p15", "c1", 0, "c0", 1): ACTLR,
        ("p15", "c1", 0, "c0", 2): CPACR,

        ("p15", "c1", 0, "c1", 0): SCR,
        ("p15", "c1", 0, "c1", 1): SDER,
        ("p15", "c1", 0, "c1", 2): NSACR,

        ("p15", "c1", 4, "c0", 0): HSCTLR,
        ("p15", "c1", 4, "c0", 1): HACTLR,

        ("p15", "c1", 4, "c1", 0): HCR,
        ("p15", "c1", 4, "c1", 1): HDCR,
        ("p15", "c1", 4, "c1", 2): HCPTR,
        ("p15", "c1", 4, "c1", 3): HSTR,
        ("p15", "c1", 4, "c1", 7): HACR,

        # TODO: TTBRO/TTBR1 64-bit
        ("p15", "c2", 0, "c0", 0): TTBR0,
        ("p15", "c2", 0, "c0", 1): TTBR1,
        ("p15", "c2", 0, "c0", 2): TTBCR,

        ("p15", "c2", 4, "c0", 2): HTCR,

        ("p15", "c2", 4, "c1", 2): VTCR,

        # TODO: HTTBR, VTTBR

        ("p15", "c3", 0, "c0", 0): DACR,

        ("p15", "c5", 0, "c0", 0): DFSR,
        ("p15", "c5", 0, "c0", 1): IFSR,

        ("p15", "c5", 0, "c1", 0): ADFSR,
        ("p15", "c5", 0, "c1", 1): AIFSR,

        ("p15", "c5", 4, "c1", 0): HADFSR,
        ("p15", "c5", 4, "c1", 1): HAIFSR,

        ("p15", "c5", 4, "c2", 0): HSR,

        ("p15", "c6", 0, "c1", 0): DFAR,
        ("p15", "c6", 0, "c1", 2): IFAR,

        ("p15", "c6", 4, "c0", 0): HDFAR,
        ("p15", "c6", 4, "c0", 2): HIFAR,
        ("p15", "c6", 4, "c0", 4): HPFAR,

        ("p15", "c7", 0, "c1", 0): ICIALLUIS,
        ("p15", "c7", 0, "c1", 6): BPIALLIS,

        ("p15", "c7", 0, "c4", 0): PAR,

        # TODO: PAR 64-bit

        ("p15", "c7", 0, "c5", 0): ICIALLU,
        ("p15", "c7", 0, "c5", 1): ICIMVAU,
        ("p15", "c7", 0, "c5", 4): CP15ISB,
        ("p15", "c7", 0, "c5", 6): BPIALL,
        ("p15", "c7", 0, "c5", 7): BPIMVA,

        ("p15", "c7", 0, "c6", 1): DCIMVAC,
        ("p15", "c7", 0, "c6", 2): DCISW,

        ("p15", "c7", 0, "c8", 0): ATS1CPR,
        ("p15", "c7", 0, "c8", 1): ATS1CPW,
        ("p15", "c7", 0, "c8", 2): ATS1CUR,
        ("p15", "c7", 0, "c8", 3): ATS1CUW,
        ("p15", "c7", 0, "c8", 4): ATS12NSOPR,
        ("p15", "c7", 0, "c8", 5): ATS12NSOPW,
        ("p15", "c7", 0, "c8", 6): ATS12NSOUR,
        ("p15", "c7", 0, "c8", 7): ATS12NSOUW,

        ("p15", "c7", 0, "c10", 1): DCCMVAC,
        ("p15", "c7", 0, "c10", 2): DCCSW,
        ("p15", "c7", 0, "c10", 4): CP15DSB,
        ("p15", "c7", 0, "c10", 5): CP15DMB,

        ("p15", "c7", 0, "c11", 1): DCCMVAU,

        ("p15", "c7", 0, "c14", 1): DCCIMVAC,
        ("p15", "c7", 0, "c14", 2): DCCISW,

        ("p15", "c7", 4, "c8", 0): ATS1HR,
        ("p15", "c7", 4, "c8", 1): ATS1HW,

        ("p15", "c8", 0, "c3", 0): TLBIALLIS,
        ("p15", "c8", 0, "c3", 1): TLBIMVAIS,
        ("p15", "c8", 0, "c3", 2): TLBIASIDIS,
        ("p15", "c8", 0, "c3", 3): TLBIMVAAIS,

        ("p15", "c8", 0, "c5", 0): ITLBIALL,
        ("p15", "c8", 0, "c5", 1): ITLBIMVA,
        ("p15", "c8", 0, "c5", 2): ITLBIASID,

        ("p15", "c8", 0, "c6", 0): DTLBIALL,
        ("p15", "c8", 0, "c6", 1): DTLBIMVA,
        ("p15", "c8", 0, "c6", 2): DTLBIASID,

        ("p15", "c8", 0, "c7", 0): TLBIALL,
        ("p15", "c8", 0, "c7", 1): TLBIMVA,
        ("p15", "c8", 0, "c7", 2): TLBIASID,
        ("p15", "c8", 0, "c7", 3): TLBIMVAA,

        ("p15", "c8", 4, "c3", 0): TLBIALLHIS,
        ("p15", "c8", 4, "c3", 1): TLBIMVAHIS,
        ("p15", "c8", 4, "c3", 4): TLBIALLNSNHIS,

        ("p15", "c8", 4, "c7", 0): TLBIALLH,
        ("p15", "c8", 4, "c7", 1): TLBIMVAH,
        ("p15", "c8", 4, "c7", 2): TLBIALLNSNH,

        ("p15", "c9", 0, "c12", 0): PMCR,
        ("p15", "c9", 0, "c12", 1): PMCNTENSET,
        ("p15", "c9", 0, "c12", 2): PMCNTENCLR,
        ("p15", "c9", 0, "c12", 3): PMOVSR,
        ("p15", "c9", 0, "c12", 4): PMSWINC,
        ("p15", "c9", 0, "c12", 5): PMSELR,
        ("p15", "c9", 0, "c12", 6): PMCEID0,
        ("p15", "c9", 0, "c12", 7): PMCEID1,

        ("p15", "c9", 0, "c13", 0): PMCCNTR,
        ("p15", "c9", 0, "c13", 1): PMXEVTYPER,
        ("p15", "c9", 0, "c13", 2): PMXEVCNTR,

        ("p15", "c9", 0, "c14", 0): PMUSERENR,
        ("p15", "c9", 0, "c14", 1): PMINTENSET,
        ("p15", "c9", 0, "c14", 2): PMINTENCLR,
        ("p15", "c9", 0, "c14", 3): PMOVSSET,

        ("p15", "c10", 0, "c2", 0): PRRR,   # ALIAS MAIR0
        ("p15", "c10", 0, "c2", 1): NMRR,   # ALIAS MAIR1

        ("p15", "c10", 0, "c3", 0): AMAIR0,
        ("p15", "c10", 0, "c3", 1): AMAIR1,

        ("p15", "c10", 4, "c2", 0): HMAIR0,
        ("p15", "c10", 4, "c2", 1): HMAIR1,

        ("p15", "c10", 4, "c3", 0): HAMAIR0,
        ("p15", "c10", 4, "c3", 1): HAMAIR1,

        ("p15", "c12", 0, "c0", 0): VBAR,
        ("p15", "c12", 0, "c0", 1): MVBAR,

        ("p15", "c12", 0, "c1", 0): ISR,

        ("p15", "c12", 4, "c0", 0): HVBAR,

        ("p15", "c13", 0, "c0", 0): FCSEIDR,
        ("p15", "c13", 0, "c0", 1): CONTEXTIDR,
        ("p15", "c13", 0, "c0", 2): TPIDRURW,
        ("p15", "c13", 0, "c0", 3): TPIDRURO,
        ("p15", "c13", 0, "c0", 4): TPIDRPRW,

        ("p15", "c13", 4, "c0", 2): HTPIDR,

        ("p15", "c14", 0, "c0", 0): CNTFRQ,
        # TODO: CNTPCT 64-bit

        ("p15", "c14", 0, "c1", 0): CNTKCTL,

        ("p15", "c14", 0, "c2", 0): CNTP_TVAL,
        ("p15", "c14", 0, "c2", 1): CNTP_CTL,

        ("p15", "c14", 0, "c3", 0): CNTV_TVAL,
        ("p15", "c14", 0, "c3", 1): CNTV_CTL,

        # TODO: CNTVCT, CNTP_CVAL, CNTV_CVAL, CNTVOFF 64-bit

        ("p15", "c14", 4, "c1", 0): CNTHCTL,

        ("p15", "c14", 4, "c2", 0): CNTHP_TVAL,
        ("p15", "c14", 4, "c2", 0): CNTHP_CTL

        # TODO: CNTHP_CVAL 64-bit
        }

# liris.cnrs.fr/~mmrissa/lib/exe/fetch.php?media=armv7-a-r-manual.pdf
EXCEPT_SOFT_BP = (1 << 1)

EXCEPT_PRIV_INSN = (1 << 17)

# CPSR: N Z C V


def update_flag_zf(a):
    return [ExprAssign(zf, ExprOp("FLAG_EQ", a))]


def update_flag_zf_eq(a, b):
    return [ExprAssign(zf, ExprOp("FLAG_EQ_CMP", a, b))]


def update_flag_nf(arg):
    return [
        ExprAssign(
            nf,
            ExprOp("FLAG_SIGN_SUB", arg, ExprInt(0, arg.size))
        )
    ]


def update_flag_zn(a):
    e = []
    e += update_flag_zf(a)
    e += update_flag_nf(a)
    return e



# XXX TODO: set cf if ROT imm in argument


def check_ops_msb(a, b, c):
    if not a or not b or not c or a != b or a != c:
        raise ValueError('bad ops size %s %s %s' % (a, b, c))

def update_flag_add_cf(op1, op2):
    "Compute cf in @op1 + @op2"
    return [ExprAssign(cf, ExprOp("FLAG_ADD_CF", op1, op2))]


def update_flag_add_of(op1, op2):
    "Compute of in @op1 + @op2"
    return [ExprAssign(of, ExprOp("FLAG_ADD_OF", op1, op2))]


def update_flag_sub_cf(op1, op2):
    "Compote CF in @op1 - @op2"
    return [ExprAssign(cf, ExprOp("FLAG_SUB_CF", op1, op2) ^ ExprInt(1, 1))]


def update_flag_sub_of(op1, op2):
    "Compote OF in @op1 - @op2"
    return [ExprAssign(of, ExprOp("FLAG_SUB_OF", op1, op2))]


def update_flag_arith_add_co(arg1, arg2):
    e = []
    e += update_flag_add_cf(arg1, arg2)
    e += update_flag_add_of(arg1, arg2)
    return e


def update_flag_arith_add_zn(arg1, arg2):
    """
    Compute zf and nf flags for (arg1 + arg2)
    """
    e = []
    e += update_flag_zf_eq(arg1, -arg2)
    e += [ExprAssign(nf, ExprOp("FLAG_SIGN_SUB", arg1, -arg2))]
    return e


def update_flag_arith_sub_co(arg1, arg2):
    """
    Compute cf and of flags for (arg1 - arg2)
    """
    e = []
    e += update_flag_sub_cf(arg1, arg2)
    e += update_flag_sub_of(arg1, arg2)
    return e


def update_flag_arith_sub_zn(arg1, arg2):
    """
    Compute zf and nf flags for (arg1 - arg2)
    """
    e = []
    e += update_flag_zf_eq(arg1, arg2)
    e += [ExprAssign(nf, ExprOp("FLAG_SIGN_SUB", arg1, arg2))]
    return e




def update_flag_zfaddwc_eq(arg1, arg2, arg3):
    return [ExprAssign(zf, ExprOp("FLAG_EQ_ADDWC", arg1, arg2, arg3))]

def update_flag_zfsubwc_eq(arg1, arg2, arg3):
    return [ExprAssign(zf, ExprOp("FLAG_EQ_SUBWC", arg1, arg2, arg3))]


def update_flag_arith_addwc_zn(arg1, arg2, arg3):
    """
    Compute znp flags for (arg1 + arg2 + cf)
    """
    e = []
    e += update_flag_zfaddwc_eq(arg1, arg2, arg3)
    e += [ExprAssign(nf, ExprOp("FLAG_SIGN_ADDWC", arg1, arg2, arg3))]
    return e


def update_flag_arith_subwc_zn(arg1, arg2, arg3):
    """
    Compute znp flags for (arg1 - (arg2 + cf))
    """
    e = []
    e += update_flag_zfsubwc_eq(arg1, arg2, arg3)
    e += [ExprAssign(nf, ExprOp("FLAG_SIGN_SUBWC", arg1, arg2, arg3))]
    return e


def update_flag_addwc_cf(op1, op2, op3):
    "Compute cf in @res = @op1 + @op2 + @op3"
    return [ExprAssign(cf, ExprOp("FLAG_ADDWC_CF", op1, op2, op3))]


def update_flag_addwc_of(op1, op2, op3):
    "Compute of in @res = @op1 + @op2 + @op3"
    return [ExprAssign(of, ExprOp("FLAG_ADDWC_OF", op1, op2, op3))]


def update_flag_arith_addwc_co(arg1, arg2, arg3):
    e = []
    e += update_flag_addwc_cf(arg1, arg2, arg3)
    e += update_flag_addwc_of(arg1, arg2, arg3)
    return e



def update_flag_subwc_cf(op1, op2, op3):
    "Compute cf in @res = @op1 + @op2 + @op3"
    return [ExprAssign(cf, ExprOp("FLAG_SUBWC_CF", op1, op2, op3) ^ ExprInt(1, 1))]


def update_flag_subwc_of(op1, op2, op3):
    "Compute of in @res = @op1 + @op2 + @op3"
    return [ExprAssign(of, ExprOp("FLAG_SUBWC_OF", op1, op2, op3))]


def update_flag_arith_subwc_co(arg1, arg2, arg3):
    e = []
    e += update_flag_subwc_cf(arg1, arg2, arg3)
    e += update_flag_subwc_of(arg1, arg2, arg3)
    return e



def get_dst(a):
    if a == PC:
        return PC
    return None

# instruction definition ##############


def adc(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    r = b + c + cf.zeroExtend(32)
    if instr.name == 'ADCS' and a != PC:
        e += update_flag_arith_addwc_zn(arg1, arg2, cf)
        e += update_flag_arith_addwc_co(arg1, arg2, cf)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def add(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    r = b + c
    if instr.name == 'ADDS' and a != PC:
        e += update_flag_arith_add_zn(arg1, arg2)
        e += update_flag_arith_add_co(arg1, arg2)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def l_and(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b & c
    if instr.name == 'ANDS' and a != PC:
        e += [ExprAssign(zf, ExprOp('FLAG_EQ_AND', b, c))]
        e += update_flag_nf(r)

    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def sub(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b - c
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def subs(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    r = b - c
    e += update_flag_arith_sub_zn(arg1, arg2)
    e += update_flag_arith_sub_co(arg1, arg2)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def eor(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b ^ c
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def eors(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    r = arg1 ^ arg2

    e += [ExprAssign(zf, ExprOp('FLAG_EQ_CMP', arg1, arg2))]
    e += update_flag_nf(r)

    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def rsb(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = c, b
    r = arg1 - arg2
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def rsbs(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = c, b
    r = arg1 - arg2
    e += update_flag_arith_sub_zn(arg1, arg2)
    e += update_flag_arith_sub_co(arg1, arg2)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def sbc(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    r = arg1 - (arg2 + (~cf).zeroExtend(32))
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def sbcs(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    r = arg1 - (arg2 + (~cf).zeroExtend(32))

    e += update_flag_arith_subwc_zn(arg1, arg2, ~cf)
    e += update_flag_arith_subwc_co(arg1, arg2, ~cf)

    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def rsc(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = c, b
    r = arg1 - (arg2 + (~cf).zeroExtend(32))
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def rscs(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = c, b
    r = arg1 - (arg2 + (~cf).zeroExtend(32))

    e += update_flag_arith_subwc_zn(arg1, arg2, ~cf)
    e += update_flag_arith_subwc_co(arg1, arg2, ~cf)

    e.append(ExprAssign(a, r))

    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def tst(ir, instr, a, b):
    e = []
    arg1, arg2 = a, b
    r = arg1 & arg2

    e += [ExprAssign(zf, ExprOp('FLAG_EQ_AND', arg1, arg2))]
    e += update_flag_nf(r)

    return e, []


def teq(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    r = arg1 ^ arg2

    e += [ExprAssign(zf, ExprOp('FLAG_EQ_CMP', arg1, arg2))]
    e += update_flag_nf(r)

    return e, []


def l_cmp(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    e += update_flag_arith_sub_zn(arg1, arg2)
    e += update_flag_arith_sub_co(arg1, arg2)
    return e, []


def cmn(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    e += update_flag_arith_add_zn(arg1, arg2)
    e += update_flag_arith_add_co(arg1, arg2)
    return e, []


def orr(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b | c
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def orn(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = ~(b | c)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def orrs(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    arg1, arg2 = b, c
    r = arg1 | arg2

    e += [ExprAssign(zf, ExprOp('FLAG_EQ', r))]
    e += update_flag_nf(r)

    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def mov(ir, instr, a, b):
    e = [ExprAssign(a, b)]
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, b))
    return e, []


def movt(ir, instr, a, b):
    r = a | b << ExprInt(16, 32)
    e = [ExprAssign(a, r)]
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def movs(ir, instr, a, b):
    e = []
    e.append(ExprAssign(a, b))
    # XXX TODO check
    e += [ExprAssign(zf, ExprOp('FLAG_EQ', b))]
    e += update_flag_nf(b)

    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, b))
    return e, []


def mvn(ir, instr, a, b):
    r = b ^ ExprInt(-1, 32)
    e = [ExprAssign(a, r)]
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def mvns(ir, instr, a, b):
    e = []
    r = b ^ ExprInt(-1, 32)
    e.append(ExprAssign(a, r))
    # XXX TODO check
    e += [ExprAssign(zf, ExprOp('FLAG_EQ', r))]
    e += update_flag_nf(r)

    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []



def mrs(ir, instr, a, b):
    e = []
    if b.is_id('CPSR_cxsf'):
        out = []
        out.append(ExprInt(0x10, 28))
        out.append(of)
        out.append(cf)
        out.append(zf)
        out.append(nf)
        e.append(ExprAssign(a, ExprCompose(*out)))
    else:
        raise NotImplementedError("MRS not implemented")
    return e, []

def msr(ir, instr, a, b):
    e = []
    if a.is_id('CPSR_cf'):
        e.append(ExprAssign(nf, b[31:32]))
        e.append(ExprAssign(zf, b[30:31]))
        e.append(ExprAssign(cf, b[29:30]))
        e.append(ExprAssign(of, b[28:29]))
    else:
        raise NotImplementedError("MSR not implemented")
    return e, []


def neg(ir, instr, a, b):
    e = []
    r = - b
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def negs(ir, instr, a, b):
    return subs(ir, instr, a, ExprInt(0, b.size), b)

def bic(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b & (c ^ ExprInt(-1, 32))
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def bics(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    tmp1, tmp2 = b, ~c
    r = tmp1 & tmp2

    e += [ExprAssign(zf, ExprOp('FLAG_EQ_AND', tmp1, tmp2))]
    e += update_flag_nf(r)

    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def sdiv(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b

    loc_div = ExprLoc(ir.loc_db.add_location(), ir.IRDst.size)
    loc_except = ExprId(ir.loc_db.add_location(), ir.IRDst.size)
    loc_next = ExprLoc(ir.get_next_loc_key(instr), ir.IRDst.size)

    e.append(ExprAssign(ir.IRDst, ExprCond(c, loc_div, loc_except)))

    do_except = []
    do_except.append(ExprAssign(exception_flags, ExprInt(EXCEPT_DIV_BY_ZERO, exception_flags.size)))
    do_except.append(ExprAssign(ir.IRDst, loc_next))
    blk_except = IRBlock(ir.loc_db, loc_except.loc_key, [AssignBlock(do_except, instr)])



    r = ExprOp("sdiv", b, c)
    do_div = []
    do_div.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        do_div.append(ExprAssign(ir.IRDst, r))

    do_div.append(ExprAssign(ir.IRDst, loc_next))
    blk_div = IRBlock(ir.loc_db, loc_div.loc_key, [AssignBlock(do_div, instr)])

    return e, [blk_div, blk_except]


def udiv(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b



    loc_div = ExprLoc(ir.loc_db.add_location(), ir.IRDst.size)
    loc_except = ExprLoc(ir.loc_db.add_location(), ir.IRDst.size)
    loc_next = ExprLoc(ir.get_next_loc_key(instr), ir.IRDst.size)

    e.append(ExprAssign(ir.IRDst, ExprCond(c, loc_div, loc_except)))

    do_except = []
    do_except.append(ExprAssign(exception_flags, ExprInt(EXCEPT_DIV_BY_ZERO, exception_flags.size)))
    do_except.append(ExprAssign(ir.IRDst, loc_next))
    blk_except = IRBlock(ir.loc_db, loc_except.loc_key, [AssignBlock(do_except, instr)])


    r = ExprOp("udiv", b, c)
    do_div = []
    do_div.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        do_div.append(ExprAssign(ir.IRDst, r))

    do_div.append(ExprAssign(ir.IRDst, loc_next))
    blk_div = IRBlock(ir.loc_db, loc_div.loc_key, [AssignBlock(do_div, instr)])

    return e, [blk_div, blk_except]


def mla(ir, instr, a, b, c, d):
    e = []
    r = (b * c) + d
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def mlas(ir, instr, a, b, c, d):
    e = []
    r = (b * c) + d
    e += update_flag_zn(r)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def mls(ir, instr, a, b, c, d):
    e = []
    r = d - (b * c)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def mul(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b * c
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def muls(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b * c
    e += update_flag_zn(r)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def umull(ir, instr, a, b, c, d):
    e = []
    r = c.zeroExtend(64) * d.zeroExtend(64)
    e.append(ExprAssign(a, r[0:32]))
    e.append(ExprAssign(b, r[32:64]))
    # r15/IRDst not allowed as output
    return e, []

def umlal(ir, instr, a, b, c, d):
    e = []
    r = c.zeroExtend(64) * d.zeroExtend(64) + ExprCompose(a, b)
    e.append(ExprAssign(a, r[0:32]))
    e.append(ExprAssign(b, r[32:64]))
    # r15/IRDst not allowed as output
    return e, []

def smull(ir, instr, a, b, c, d):
    e = []
    r = c.signExtend(64) * d.signExtend(64)
    e.append(ExprAssign(a, r[0:32]))
    e.append(ExprAssign(b, r[32:64]))
    # r15/IRDst not allowed as output
    return e, []

def smlal(ir, instr, a, b, c, d):
    e = []
    r = c.signExtend(64) * d.signExtend(64) + ExprCompose(a, b)
    e.append(ExprAssign(a, r[0:32]))
    e.append(ExprAssign(b, r[32:64]))
    # r15/IRDst not allowed as output
    return e, []

def b(ir, instr, a):
    e = []
    e.append(ExprAssign(PC, a))
    e.append(ExprAssign(ir.IRDst, a))
    return e, []


def bl(ir, instr, a):
    e = []
    l = ExprInt(instr.offset + instr.l, 32)
    e.append(ExprAssign(PC, a))
    e.append(ExprAssign(ir.IRDst, a))
    e.append(ExprAssign(LR, l))
    return e, []


def bx(ir, instr, a):
    e = []
    e.append(ExprAssign(PC, a))
    e.append(ExprAssign(ir.IRDst, a))
    return e, []


def blx(ir, instr, a):
    e = []
    l = ExprInt(instr.offset + instr.l, 32)
    e.append(ExprAssign(PC, a))
    e.append(ExprAssign(ir.IRDst, a))
    e.append(ExprAssign(LR, l))
    return e, []


def st_ld_r(ir, instr, a, a2, b, store=False, size=32, s_ext=False, z_ext=False):
    e = []
    wb = False
    postinc = False
    b = b.ptr
    if isinstance(b, ExprOp):
        if b.op == "wback":
            wb = True
            b = b.args[0]
        if b.op == "postinc":
            postinc = True
    if isinstance(b, ExprOp) and b.op in ["postinc", 'preinc']:
        # XXX TODO CHECK
        base, off = b.args[0],  b.args[1]  # ExprInt(size/8, 32)
    else:
        base, off = b, ExprInt(0, 32)
    if postinc:
        ad = base
    else:
        ad = base + off

    # PC base lookup uses PC 4 byte alignment
    ad = ad.replace_expr({PC: PC & ExprInt(0xFFFFFFFC, 32)})

    dmem = False
    if size in [8, 16]:
        if store:
            a = a[:size]
            m = ExprMem(ad, size=size)
        elif s_ext:
            m = ExprMem(ad, size=size).signExtend(a.size)
        elif z_ext:
            m = ExprMem(ad, size=size).zeroExtend(a.size)
        else:
            raise ValueError('unhandled case')
    elif size == 32:
        m = ExprMem(ad, size=size)
    elif size == 64:
        assert a2 is not None
        m = ExprMem(ad, size=32)
        dmem = True
        size = 32
    else:
        raise ValueError('the size DOES matter')
    dst = None

    if store:
        e.append(ExprAssign(m, a))
        if dmem:
            e.append(ExprAssign(ExprMem(ad + ExprInt(4, 32), size=size), a2))
    else:
        if a == PC:
            dst = PC
            e.append(ExprAssign(ir.IRDst, m))
        e.append(ExprAssign(a, m))
        if dmem:
            e.append(ExprAssign(a2, ExprMem(ad + ExprInt(4, 32), size=size)))

    # XXX TODO check multiple write cause by wb
    if wb or postinc:
        e.append(ExprAssign(base, base + off))
    return e, []


def ldr(ir, instr, a, b):
    return st_ld_r(ir, instr, a, None, b, store=False)


def ldrd(ir, instr, a, b, c=None):
    if c is None:
        a2 = ir.arch.regs.all_regs_ids[ir.arch.regs.all_regs_ids.index(a) + 1]
    else:
        a2 = b
        b = c
    return st_ld_r(ir, instr, a, a2, b, store=False, size=64)


def l_str(ir, instr, a, b):
    return st_ld_r(ir, instr, a, None, b, store=True)


def l_strd(ir, instr, a, b, c=None):
    if c is None:
        a2 = ir.arch.regs.all_regs_ids[ir.arch.regs.all_regs_ids.index(a) + 1]
    else:
        a2 = b
        b = c
    return st_ld_r(ir, instr, a, a2, b, store=True, size=64)

def ldrb(ir, instr, a, b):
    return st_ld_r(ir, instr, a, None, b, store=False, size=8, z_ext=True)

def ldrsb(ir, instr, a, b):
    return st_ld_r(ir, instr, a, None, b, store=False, size=8, s_ext=True, z_ext=False)

def strb(ir, instr, a, b):
    return st_ld_r(ir, instr, a, None, b, store=True, size=8)

def ldrh(ir, instr, a, b):
    return st_ld_r(ir, instr, a, None, b, store=False, size=16, z_ext=True)


def strh(ir, instr, a, b):
    return st_ld_r(ir, instr, a, None, b, store=True, size=16, z_ext=True)


def ldrsh(ir, instr, a, b):
    return st_ld_r(ir, instr, a, None, b, store=False, size=16, s_ext=True, z_ext=False)


def st_ld_m(ir, instr, a, b, store=False, postinc=False, updown=False):
    e = []
    wb = False
    dst = None
    if isinstance(a, ExprOp) and a.op == 'wback':
        wb = True
        a = a.args[0]
    if isinstance(b, ExprOp) and b.op == 'sbit':
        b = b.args[0]
    regs = b.args
    base = a
    if updown:
        step = 4
    else:
        step = -4
        regs = regs[::-1]
    if postinc:
        pass
    else:
        base += ExprInt(step, 32)
    for i, r in enumerate(regs):
        ad = base + ExprInt(i * step, 32)
        if store:
            e.append(ExprAssign(ExprMem(ad, 32), r))
        else:
            e.append(ExprAssign(r, ExprMem(ad, 32)))
            if r == PC:
                e.append(ExprAssign(ir.IRDst, ExprMem(ad, 32)))
    # XXX TODO check multiple write cause by wb
    if wb:
        if postinc:
            e.append(ExprAssign(a, base + ExprInt(len(regs) * step, 32)))
        else:
            e.append(ExprAssign(a, base + ExprInt((len(regs) - 1) * step, 32)))
    if store:
        pass
    else:
        assert(isinstance(b, ExprOp) and b.op == "reglist")

    return e, []


def ldmia(ir, instr, a, b):
    return st_ld_m(ir, instr, a, b, store=False, postinc=True, updown=True)


def ldmib(ir, instr, a, b):
    return st_ld_m(ir, instr, a, b, store=False, postinc=False, updown=True)


def ldmda(ir, instr, a, b):
    return st_ld_m(ir, instr, a, b, store=False, postinc=True, updown=False)


def ldmdb(ir, instr, a, b):
    return st_ld_m(ir, instr, a, b, store=False, postinc=False, updown=False)


def stmia(ir, instr, a, b):
    return st_ld_m(ir, instr, a, b, store=True, postinc=True, updown=True)


def stmib(ir, instr, a, b):
    return st_ld_m(ir, instr, a, b, store=True, postinc=False, updown=True)


def stmda(ir, instr, a, b):
    return st_ld_m(ir, instr, a, b, store=True, postinc=True, updown=False)


def stmdb(ir, instr, a, b):
    return st_ld_m(ir, instr, a, b, store=True, postinc=False, updown=False)


def svc(ir, instr, a):
    e = []
    except_int = EXCEPT_INT_XX
    e.append(ExprAssign(exception_flags, ExprInt(except_int, 32)))
    e.append(ExprAssign(interrupt_num, a))
    return e, []


def und(ir, instr, a, b):
    # XXX TODO implement
    e = []
    return e, []

# TODO XXX implement correct CF for shifters
def lsr(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b >> c
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def lsrs(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b >> c
    e.append(ExprAssign(a, r))

    e += [ExprAssign(zf, ExprOp('FLAG_EQ', r))]
    e += update_flag_nf(r)

    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def asr(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = ExprOp("a>>", b, c)
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def asrs(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = ExprOp("a>>", b, c)
    e.append(ExprAssign(a, r))

    e += [ExprAssign(zf, ExprOp('FLAG_EQ', r))]
    e += update_flag_nf(r)

    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def lsl(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b << c
    e.append(ExprAssign(a, r))
    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def lsls(ir, instr, a, b, c=None):
    e = []
    if c is None:
        b, c = a, b
    r = b << c
    e.append(ExprAssign(a, r))

    e += [ExprAssign(zf, ExprOp('FLAG_EQ', r))]
    e += update_flag_nf(r)

    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def rors(ir, instr, a, b):
    e = []
    r = ExprOp(">>>", a, b)
    e.append(ExprAssign(a, r))

    e += [ExprAssign(zf, ExprOp('FLAG_EQ', r))]
    e += update_flag_nf(r)

    dst = get_dst(a)
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def push(ir, instr, a):
    e = []
    regs = list(a.args)
    for i in range(len(regs)):
        r = SP + ExprInt(-4 * len(regs) + 4 * i, 32)
        e.append(ExprAssign(ExprMem(r, 32), regs[i]))
    r = SP + ExprInt(-4 * len(regs), 32)
    e.append(ExprAssign(SP, r))
    return e, []


def pop(ir, instr, a):
    e = []
    regs = list(a.args)
    dst = None
    for i in range(len(regs)):
        r = SP + ExprInt(4 * i, 32)
        e.append(ExprAssign(regs[i], ExprMem(r, 32)))
        if regs[i] == ir.pc:
            dst = ExprMem(r, 32)
    r = SP + ExprInt(4 * len(regs), 32)
    e.append(ExprAssign(SP, r))
    if dst is not None:
        e.append(ExprAssign(ir.IRDst, dst))
    return e, []


def cbz(ir, instr, a, b):
    e = []
    loc_next = ir.get_next_loc_key(instr)
    loc_next_expr = ExprLoc(loc_next, 32)
    e.append(ExprAssign(ir.IRDst, ExprCond(a, loc_next_expr, b)))
    return e, []


def cbnz(ir, instr, a, b):
    e = []
    loc_next = ir.get_next_loc_key(instr)
    loc_next_expr = ExprLoc(loc_next, 32)
    e.append(ExprAssign(ir.IRDst, ExprCond(a, b, loc_next_expr)))
    return e, []


def uxtb(ir, instr, a, b):
    e = []
    r = b[:8].zeroExtend(32)
    e.append(ExprAssign(a, r))
    dst = None
    if PC in a.get_r():
        dst = PC
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def uxth(ir, instr, a, b):
    e = []
    r = b[:16].zeroExtend(32)
    e.append(ExprAssign(a, r))
    dst = None
    if PC in a.get_r():
        dst = PC
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def sxtb(ir, instr, a, b):
    e = []
    r = b[:8].signExtend(32)
    e.append(ExprAssign(a, r))
    dst = None
    if PC in a.get_r():
        dst = PC
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def sxth(ir, instr, a, b):
    e = []
    r = b[:16].signExtend(32)
    e.append(ExprAssign(a, r))
    dst = None
    if PC in a.get_r():
        dst = PC
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def ubfx(ir, instr, a, b, c, d):
    e = []
    c = int(c)
    d = int(d)
    r = b[c:c+d].zeroExtend(32)
    e.append(ExprAssign(a, r))
    dst = None
    if PC in a.get_r():
        dst = PC
        e.append(ExprAssign(ir.IRDst, r))
    return e, []

def bfc(ir, instr, a, b, c):
    e = []
    start = int(b)
    stop = start + int(c)
    out = []
    last = 0
    if start:
        out.append(a[:start])
        last = start
    if stop - start:
        out.append(ExprInt(0, 32)[last:stop])
        last = stop
    if last < 32:
        out.append(a[last:])
    r = ExprCompose(*out)
    e.append(ExprAssign(a, r))
    dst = None
    if PC in a.get_r():
        dst = PC
        e.append(ExprAssign(ir.IRDst, r))
    return e, []


def pld(ir, instr, a):
    e = []
    return e, []


def pldw(ir, instr, a):
    e = []
    return e, []


def clz(ir, instr, a, b):
    e = []
    e.append(ExprAssign(a, ExprOp('cntleadzeros', b)))
    return e, []

def uxtab(ir, instr, a, b, c):
    e = []
    e.append(ExprAssign(a, b + (c & ExprInt(0xff, 32))))
    return e, []


def uxtah(ir, instr, a, b, c):
    e = []
    e.append(ExprAssign(a, b + (c & ExprInt(0xffff, 32))))
    return e, []


def bkpt(ir, instr, a):
    e = []
    e.append(ExprAssign(exception_flags, ExprInt(EXCEPT_SOFT_BP, 32)))
    e.append(ExprAssign(bp_num, a))
    return e, []


def _extract_s16(arg, part):
    if part == 'B': # bottom 16 bits
        return arg[0:16]
    elif part == 'T': # top 16 bits
        return arg[16:32]


def smul(ir, instr, a, b, c):
    e = []
    e.append(ExprAssign(a, _extract_s16(b, instr.name[4]).signExtend(32) * _extract_s16(c, instr.name[5]).signExtend(32)))
    return e, []


def smulw(ir, instr, a, b, c):
    e = []
    prod = b.signExtend(48) * _extract_s16(c, instr.name[5]).signExtend(48)
    e.append(ExprAssign(a, prod[16:48]))
    return e, [] # signed most significant 32 bits of the 48-bit result


def tbb(ir, instr, a):
    e = []
    dst = PC + ExprInt(2, 32) * a.zeroExtend(32)
    e.append(ExprAssign(PC, dst))
    e.append(ExprAssign(ir.IRDst, dst))
    return e, []


def tbh(ir, instr, a):
    e = []
    dst = PC + ExprInt(2, 32) * a.zeroExtend(32)
    e.append(ExprAssign(PC, dst))
    e.append(ExprAssign(ir.IRDst, dst))
    return e, []


def smlabb(ir, instr, a, b, c, d):
    e = []
    result = (b[:16].signExtend(32) * c[:16].signExtend(32)) + d
    e.append(ExprAssign(a, result))
    return e, []


def smlabt(ir, instr, a, b, c, d):
    e = []
    result = (b[:16].signExtend(32) * c[16:32].signExtend(32)) + d
    e.append(ExprAssign(a, result))
    return e, []


def smlatb(ir, instr, a, b, c, d):
    e = []
    result = (b[16:32].signExtend(32) * c[:16].signExtend(32)) + d
    e.append(ExprAssign(a, result))
    return e, []


def smlatt(ir, instr, a, b, c, d):
    e = []
    result = (b[16:32].signExtend(32) * c[16:32].signExtend(32)) + d
    e.append(ExprAssign(a, result))
    return e, []


def uadd8(ir, instr, a, b, c):
    e = []
    sums = []
    ges = []
    for i in range(0, 32, 8):
        sums.append(b[i:i+8] + c[i:i+8])
        ges.append((b[i:i+8].zeroExtend(9) + c[i:i+8].zeroExtend(9))[8:9])

    e.append(ExprAssign(a, ExprCompose(*sums)))

    for i, value in enumerate(ges):
        e.append(ExprAssign(ge_regs[i], value))
    return e, []


def sel(ir, instr, a, b, c):
    e = []
    cond = nf ^ of ^ ExprInt(1, 1)
    parts = []
    for i in range(4):
        parts.append(ExprCond(ge_regs[i], b[i*8:(i+1)*8], c[i*8:(i+1)*8]))
    result = ExprCompose(*parts)
    e.append(ExprAssign(a, result))
    return e, []


def rev(ir, instr, a, b):
    e = []
    result = ExprCompose(b[24:32], b[16:24], b[8:16], b[:8])
    e.append(ExprAssign(a, result))
    return e, []


def rev16(ir, instr, a, b):
    e = []
    result = ExprCompose(b[8:16], b[:8], b[24:32], b[16:24])
    e.append(ExprAssign(a, result))
    return e, []


def nop(ir, instr):
    e = []
    return e, []


def dsb(ir, instr, a):
    # XXX TODO
    e = []
    return e, []

def isb(ir, instr, a):
    # XXX TODO
    e = []
    return e, []

def cpsie(ir, instr, a):
    # XXX TODO
    e = []
    return e, []


def cpsid(ir, instr, a):
    # XXX TODO
    e = []
    return e, []


def wfe(ir, instr):
    # XXX TODO
    e = []
    return e, []


def wfi(ir, instr):
    # XXX TODO
    e = []
    return e, []

def adr(ir, instr, arg1, arg2):
    e = []
    e.append(ExprAssign(arg1, (PC & ExprInt(0xfffffffc, 32)) + arg2))
    return e, []


def pkhbt(ir, instr, arg1, arg2, arg3):
    e = []
    e.append(
        ExprAssign(
            arg1,
            ExprCompose(
                arg2[:16],
                arg3[16:]
            )
        )
    )
    return e, []


def pkhtb(ir, instr, arg1, arg2, arg3):
    e = []
    e.append(
        ExprAssign(
            arg1,
            ExprCompose(
                arg3[:16],
                arg2[16:]
            )
        )
    )
    return e, []

def mrc(ir, insr, arg1, arg2, arg3, arg4, arg5, arg6):
    e = []
    sreg = (str(arg1), str(arg4), int(arg2), str(arg5), int(arg6))
    if sreg in coproc_reg_dict:
        e.append(ExprAssign(arg3, coproc_reg_dict[sreg]))
    else:
        raise NotImplementedError("Unknown coprocessor register: %s %s %d %s %d" % (str(arg1), str(arg4), int(arg2), str(arg5), int(arg6)))

    return e, []

def mcr(ir, insr, arg1, arg2, arg3, arg4, arg5, arg6):
    e = []
    sreg = (str(arg1), str(arg4), int(arg2), str(arg5), int(arg6))
    if sreg in coproc_reg_dict:
        e.append(ExprAssign(coproc_reg_dict[sreg], arg3))
    else:
        raise NotImplementedError("Unknown coprocessor register: %s %s %d %s %d" % (str(arg1), str(arg4), int(arg2), str(arg5), int(arg6)))

    return e, []

COND_EQ = 0
COND_NE = 1
COND_CS = 2
COND_CC = 3
COND_MI = 4
COND_PL = 5
COND_VS = 6
COND_VC = 7
COND_HI = 8
COND_LS = 9
COND_GE = 10
COND_LT = 11
COND_GT = 12
COND_LE = 13
COND_AL = 14
COND_NV = 15

cond_dct = {
    COND_EQ: "EQ",
    COND_NE: "NE",
    COND_CS: "CS",
    COND_CC: "CC",
    COND_MI: "MI",
    COND_PL: "PL",
    COND_VS: "VS",
    COND_VC: "VC",
    COND_HI: "HI",
    COND_LS: "LS",
    COND_GE: "GE",
    COND_LT: "LT",
    COND_GT: "GT",
    COND_LE: "LE",
    COND_AL: "AL",
    # COND_NV: "NV",
}

cond_dct_inv = dict((name, num) for num, name in viewitems(cond_dct))


"""
Code            Meaning (for cmp or subs)                                  Flags Tested
eq              Equal.                                                     Z==1
ne              Not equal.                                                 Z==0
cs or hs        Unsigned higher or same (or carry set).                    C==1
cc or lo        Unsigned lower (or carry clear).                           C==0
mi              Negative. The mnemonic stands for "minus".                 N==1
pl              Positive or zero. The mnemonic stands for "plus".          N==0
vs              Signed overflow. The mnemonic stands for "V set".          V==1
vc              No signed overflow. The mnemonic stands for "V clear".     V==0
hi              Unsigned higher.                                           (C==1) && (Z==0)
ls              Unsigned lower or same.                                    (C==0) || (Z==1)
ge              Signed greater than or equal.                              N==V
lt              Signed less than.                                          N!=V
gt              Signed greater than.                                       (Z==0) && (N==V)
le              Signed less than or equal.                                 (Z==1) || (N!=V)
al (or omitted) Always executed.        None tested.
"""

tab_cond = {COND_EQ: ExprOp("CC_EQ", zf),
            COND_NE: ExprOp("CC_NE", zf),
            COND_CS: ExprOp("CC_U>=", cf ^ ExprInt(1, 1)), # inv cf
            COND_CC: ExprOp("CC_U<", cf ^ ExprInt(1, 1)), # inv cf
            COND_MI: ExprOp("CC_NEG", nf),
            COND_PL: ExprOp("CC_POS", nf),
            COND_VS: ExprOp("CC_sOVR", of),
            COND_VC: ExprOp("CC_sNOOVR", of),
            COND_HI: ExprOp("CC_U>", cf ^ ExprInt(1, 1), zf), # inv cf
            COND_LS: ExprOp("CC_U<=", cf ^ ExprInt(1, 1), zf), # inv cf
            COND_GE: ExprOp("CC_S>=", nf, of),
            COND_LT: ExprOp("CC_S<", nf, of),
            COND_GT: ExprOp("CC_S>", nf, of, zf),
            COND_LE: ExprOp("CC_S<=", nf, of, zf),
            }





def is_pc_written(ir, instr_ir):
    all_pc = viewvalues(ir.mn.pc)
    for ir in instr_ir:
        if ir.dst in all_pc:
            return True, ir.dst
    return False, None


def add_condition_expr(ir, instr, cond, instr_ir, extra_ir):
    if cond == COND_AL:
        return instr_ir, extra_ir
    if not cond in tab_cond:
        raise ValueError('unknown condition %r' % cond)
    cond = tab_cond[cond]



    loc_next = ir.get_next_loc_key(instr)
    loc_next_expr = ExprLoc(loc_next, 32)
    loc_do = ir.loc_db.add_location()
    loc_do_expr = ExprLoc(loc_do, 32)

    dst_cond = ExprCond(cond, loc_do_expr, loc_next_expr)
    assert(isinstance(instr_ir, list))

    has_irdst = False
    for e in instr_ir:
        if e.dst == ir.IRDst:
            has_irdst = True
            break
    if not has_irdst:
        instr_ir.append(ExprAssign(ir.IRDst, loc_next_expr))
    e_do = IRBlock(ir.loc_db, loc_do, [AssignBlock(instr_ir, instr)])
    e = [ExprAssign(ir.IRDst, dst_cond)]
    return e, [e_do] + extra_ir

mnemo_func = {}
mnemo_func_cond = {}
mnemo_condm0 = {'add': add,
                'sub': sub,
                'eor': eor,
                'and': l_and,
                'rsb': rsb,
                'adc': adc,
                'sbc': sbc,
                'rsc': rsc,

                'tst': tst,
                'teq': teq,
                'cmp': l_cmp,
                'cmn': cmn,
                'orr': orr,
                'mov': mov,
                'movt': movt,
                'bic': bic,
                'mvn': mvn,
                'neg': neg,

                'sdiv': sdiv,
                'udiv': udiv,

                'mrc': mrc,
                'mcr': mcr,

                'mul': mul,
                'umull': umull,
                'umlal': umlal,
                'smull': smull,
                'smlal': smlal,
                'mla': mla,
                'ldr': ldr,
                'ldrd': ldrd,
                'ldrsb': ldrsb,
                'str': l_str,
                'strd': l_strd,
                'b': b,
                'bl': bl,
                'svc': svc,
                'und': und,
                'bx': bx,
                'ldrh': ldrh,
                'strh': strh,
                'ldrsh': ldrsh,
                'ldsh': ldrsh,
                'uxtb': uxtb,
                'uxth': uxth,
                'sxtb': sxtb,
                'sxth': sxth,
                'ubfx': ubfx,
                'bfc': bfc,
                'rev': rev,
                'rev16': rev16,
                'clz': clz,
                'uxtab': uxtab,
                'uxtah': uxtah,
                'bkpt': bkpt,
                'smulbb': smul,
                'smulbt': smul,
                'smultb': smul,
                'smultt': smul,
                'smulwt': smulw,
                'smulwb': smulw,

                'pkhtb': pkhtb,
                'pkhbt': pkhbt,

                }

mnemo_condm1 = {'adds': add,
                'subs': subs,
                'eors': eors,
                'ands': l_and,
                'rsbs': rsbs,
                'adcs': adc,
                'sbcs': sbcs,
                'rscs': rscs,

                'orrs': orrs,
                'movs': movs,
                'bics': bics,
                'mvns': mvns,

                'mrs': mrs,
                'msr': msr,

                'negs': negs,

                'muls': muls,
                'mls': mls,
                'mlas': mlas,
                'blx': blx,

                'ldrb': ldrb,
                'ldsb': ldrsb,
                'strb': strb,
                }

mnemo_condm2 = {'ldmia': ldmia,
                'ldmib': ldmib,
                'ldmda': ldmda,
                'ldmdb': ldmdb,

                'ldmfa': ldmda,
                'ldmfd': ldmia,
                'ldmea': ldmdb,
                'ldmed': ldmib,  # XXX


                'stmia': stmia,
                'stmib': stmib,
                'stmda': stmda,
                'stmdb': stmdb,

                'stmfa': stmib,
                'stmed': stmda,
                'stmfd': stmdb,
                'stmea': stmia,
                }


mnemo_nocond = {'lsr': lsr,
                'lsrs': lsrs,
                'lsl': lsl,
                'lsls': lsls,
                'rors': rors,
                'push': push,
                'pop': pop,
                'asr': asr,
                'asrs': asrs,
                'cbz': cbz,
                'cbnz': cbnz,
                'pld': pld,
                'pldw': pldw,
                'tbb': tbb,
                'tbh': tbh,
                'nop': nop,
                'dsb': dsb,
                'isb': isb,
                'cpsie': cpsie,
                'cpsid': cpsid,
                'wfe': wfe,
                'wfi': wfi,
                'adr': adr,
                'orn': orn,
                'smlabb': smlabb,
                'smlabt': smlabt,
                'smlatb': smlatb,
                'smlatt': smlatt,
                'uadd8': uadd8,
                'sel': sel,
                }

mn_cond_x = [mnemo_condm0,
             mnemo_condm1,
             mnemo_condm2]

for index, mn_base in enumerate(mn_cond_x):
    for mn, mf in viewitems(mn_base):
        for cond, cn in viewitems(cond_dct):
            if cond == COND_AL:
                cn = ""
            cn = cn.lower()
            if index == 0:
                mn_mod = mn + cn
            else:
                mn_mod = mn[:-index] + cn + mn[-index:]
            # print mn_mod
            mnemo_func_cond[mn_mod] = cond, mf

for name, mf in viewitems(mnemo_nocond):
    mnemo_func_cond[name] = COND_AL, mf


def split_expr_dst(ir, instr_ir):
    out = []
    dst = None
    for i in instr_ir:
        if i.dst == ir.pc:
            out.append(i)
            dst = ir.pc  # i.src
        else:
            out.append(i)
    return out, dst


def get_mnemo_expr(ir, instr, *args):
    if not instr.name.lower() in mnemo_func_cond:
        raise ValueError('unknown mnemo %s' % instr)
    cond, mf = mnemo_func_cond[instr.name.lower()]
    instr_ir, extra_ir = mf(ir, instr, *args)
    instr, extra_ir = add_condition_expr(ir, instr, cond, instr_ir, extra_ir)
    return instr, extra_ir

get_arm_instr_expr = get_mnemo_expr


class arminfo(object):
    mode = "arm"
    # offset


class Lifter_Arml(Lifter):
    def __init__(self, loc_db):
        Lifter.__init__(self, mn_arm, "l", loc_db)
        self.pc = PC
        self.sp = SP
        self.IRDst = ExprId('IRDst', 32)
        self.addrsize = 32



    def mod_pc(self, instr, instr_ir, extra_ir):
        # fix PC (+8 for arm)
        pc_fixed = {self.pc: ExprInt(instr.offset + 8, 32)}

        for i, expr in enumerate(instr_ir):
            dst, src = expr.dst, expr.src
            if dst != self.pc:
                dst = dst.replace_expr(pc_fixed)
            src = src.replace_expr(pc_fixed)
            instr_ir[i] = ExprAssign(dst, src)

        for idx, irblock in enumerate(extra_ir):
            extra_ir[idx] = irblock.modify_exprs(lambda expr: expr.replace_expr(pc_fixed) \
                                                 if expr != self.pc else expr,
                                                 lambda expr: expr.replace_expr(pc_fixed))

    def get_ir(self, instr):
        args = instr.args
        # ir = get_mnemo_expr(self, self.name.lower(), *args)
        if len(args) and isinstance(args[-1], ExprOp):
            if args[-1].op == 'rrx':
                args[-1] = ExprCompose(args[-1].args[0][1:], cf)
            elif (args[-1].op in ['<<', '>>', '<<a', 'a>>', '<<<', '>>>'] and
                  isinstance(args[-1].args[-1], ExprId)):
                args[-1] = ExprOp(args[-1].op,
                                  args[-1].args[0],
                                  args[-1].args[-1][:8].zeroExtend(32))
        instr_ir, extra_ir = get_mnemo_expr(self, instr, *args)

        self.mod_pc(instr, instr_ir, extra_ir)
        return instr_ir, extra_ir

    def parse_itt(self, instr):
        name = instr.name
        assert name.startswith('IT')
        name = name[1:]
        out = []
        for hint in name:
            if hint == 'T':
                out.append(0)
            elif hint == "E":
                out.append(1)
            else:
                raise ValueError("IT name invalid %s" % instr)
        return out, instr.args[0]

    def do_it_block(self, loc, index, block, assignments, gen_pc_updt):
        instr = block.lines[index]
        it_hints, it_cond = self.parse_itt(instr)
        cond_num = cond_dct_inv[it_cond.name]
        cond_eq = tab_cond[cond_num]

        if not index + len(it_hints) <= len(block.lines):
            raise NotImplementedError("Split IT block non supported yet")

        ir_blocks_all = []

        # Gen dummy irblock for IT instr
        loc_next = self.get_next_loc_key(instr)
        dst = ExprAssign(self.IRDst, ExprLoc(loc_next, 32))
        dst_blk = AssignBlock([dst], instr)
        assignments.append(dst_blk)
        irblock = IRBlock(self.loc_db, loc, assignments)
        ir_blocks_all.append([irblock])

        loc = loc_next
        assignments = []
        for hint in it_hints:
            irblocks = []
            index += 1
            instr = block.lines[index]

            # Add conditional jump to current irblock
            loc_do = self.loc_db.add_location()
            loc_next = self.get_next_loc_key(instr)

            if hint:
                local_cond = ~cond_eq
            else:
                local_cond = cond_eq
            dst = ExprAssign(self.IRDst, ExprCond(local_cond, ExprLoc(loc_do, 32), ExprLoc(loc_next, 32)))
            dst_blk = AssignBlock([dst], instr)
            assignments.append(dst_blk)
            irblock = IRBlock(self.loc_db, loc, assignments)

            irblocks.append(irblock)

            it_instr_irblocks = []
            assignments = []
            loc = loc_do

            split = self.add_instr_to_current_state(
                instr, block, assignments,
                it_instr_irblocks, gen_pc_updt
            )
            if split:
                raise NotImplementedError("Unsupported instr in IT block (%s)" % instr)

            if it_instr_irblocks:
                assert len(it_instr_irblocks) == 1
                it_instr_irblocks = it_instr_irblocks.pop()
            # Remove flags assignment if instr != [CMP, CMN, TST]
            if instr.name not in ["CMP", "CMN", "TST"]:
                # Fix assignments
                out = []
                for assignment in assignments:
                    assignment = AssignBlock(
                        {
                            dst: src for (dst, src) in viewitems(assignment)
                            if dst not in [zf, nf, of, cf]
                        },
                        assignment.instr
                    )
                    out.append(assignment)
                assignments = out
                # Fix extra irblocksx
                new_irblocks = []
                for irblock in it_instr_irblocks:
                    out = []
                    for tmp_assignment in irblock:
                        assignment = AssignBlock(
                            {
                                dst: src for (dst, src) in viewitems(assignment)
                                if dst not in [zf, nf, of, cf]
                            },
                            assignment.instr
                        )
                        out.append(assignment)
                    new_irblock = IRBlock(self.loc_db, irblock.loc_key, out)
                    new_irblocks.append(new_irblock)
                it_instr_irblocks = new_irblocks

            irblocks += it_instr_irblocks
            dst = ExprAssign(self.IRDst, ExprLoc(loc_next, 32))
            dst_blk = AssignBlock([dst], instr)
            assignments.append(dst_blk)
            irblock = IRBlock(self.loc_db, loc, assignments)
            irblocks.append(irblock)
            loc = loc_next
            assignments = []
            ir_blocks_all.append(irblocks)
        return index, ir_blocks_all

    def add_asmblock_to_ircfg(self, block, ircfg, gen_pc_updt=False):
        """
        Add a native block to the current IR
        @block: native assembly block
        @gen_pc_updt: insert PC update effects between instructions
        """

        it_hints = None
        it_cond = None
        label = block.loc_key
        assignments = []
        ir_blocks_all = []
        index = -1
        while index + 1 < len(block.lines):
            index += 1
            instr = block.lines[index]
            if label is None:
                assignments = []
                label = self.get_loc_key_for_instr(instr)
            if instr.name.startswith("IT"):
                index, irblocks_it = self.do_it_block(label, index, block, assignments, gen_pc_updt)
                for irblocks in irblocks_it:
                    ir_blocks_all += irblocks
                label = None
                continue

            split = self.add_instr_to_current_state(
                instr, block, assignments,
                ir_blocks_all, gen_pc_updt
            )
            if split:
                ir_blocks_all.append(IRBlock(self.loc_db, label, assignments))
                label = None
                assignments = []
        if label is not None:
            ir_blocks_all.append(IRBlock(self.loc_db, label, assignments))

        new_ir_blocks_all = self.post_add_asmblock_to_ircfg(block, ircfg, ir_blocks_all)
        for irblock in new_ir_blocks_all:
            ircfg.add_irblock(irblock)
        return new_ir_blocks_all



class Lifter_Armb(Lifter_Arml):
    def __init__(self, loc_db):
        Lifter.__init__(self, mn_arm, "b", loc_db)
        self.pc = PC
        self.sp = SP
        self.IRDst = ExprId('IRDst', 32)
        self.addrsize = 32


class Lifter_Armtl(Lifter_Arml):
    def __init__(self, loc_db):
        Lifter.__init__(self, mn_armt, "l", loc_db)
        self.pc = PC
        self.sp = SP
        self.IRDst = ExprId('IRDst', 32)
        self.addrsize = 32


    def mod_pc(self, instr, instr_ir, extra_ir):
        # fix PC (+4 for thumb)
        pc_fixed = {self.pc: ExprInt(instr.offset + 4, 32)}

        for i, expr in enumerate(instr_ir):
            dst, src = expr.dst, expr.src
            if dst != self.pc:
                dst = dst.replace_expr(pc_fixed)
            src = src.replace_expr(pc_fixed)
            instr_ir[i] = ExprAssign(dst, src)

        for idx, irblock in enumerate(extra_ir):
            extra_ir[idx] = irblock.modify_exprs(lambda expr: expr.replace_expr(pc_fixed) \
                                                 if expr != self.pc else expr,
                                                 lambda expr: expr.replace_expr(pc_fixed))


class Lifter_Armtb(Lifter_Armtl):
    def __init__(self, loc_db):
        Lifter.__init__(self, mn_armt, "b", loc_db)
        self.pc = PC
        self.sp = SP
        self.IRDst = ExprId('IRDst', 32)
        self.addrsize = 32




import sys
from pdb import pm
import ctypes

from miasm.analysis.binary import Container
from miasm.core.locationdb import LocationDB
from miasm.loader.pe_init import PE

ULONG = ctypes.c_uint32


def extract_409_resource(res):
    """
    Find the first resource id 0x409
    """
    if res is None:
        return None
    for x in res.resentries:
        if x.data:
            if x.name == 0x409:
                return x.data.s
        if x.offsettosubdir:
            ret = extract_409_resource(x.subdir)
            if ret:
                return ret
    return None


class ApiSetWin10(object):
    def __init__(self, fname_apisetschema):
        self.pe = PE(open(fname_apisetschema, "rb").read())
        self.version = self.get_version()

        loc_db = LocationDB()

        self.api_section = self.pe.getsectionbyname(".apiset").get_data()

    def get_version(self):
        res = self.pe.DirRes.resdesc

        data = extract_409_resource(res)
        token = b"ProductVersion\x00"
        out = []
        for c in token:
            out += [c, 0]
        out = bytes(out)
        index_start = data.find(out)
        assert index_start > 0
        index_start += len(out)
        index_stop = data.find(b"\x00\x00", index_start)
        assert index_start > 0
        data = data[index_start:index_stop]
        data = data[::2].decode()
        print("Version:", data)
        return data

    # Win 10
    """
    typedef struct _API_SET_NAMESPACE
    {
        ULONG Version;
        ULONG Size;
        ULONG Flags;
        ULONG Count;
        ULONG EntryOffset;
        ULONG HashOffset;
        ULONG HashFactor;
    } API_SET_NAMESPACE, *PAPI_SET_NAMESPACE;
    """

    class ApiSetHeader(ctypes.Structure):
        _fields_ = [
            ("Version", ULONG),
            ("Size", ULONG),
            ("Flags", ULONG),
            ("Count", ULONG),
            ("EntryOffset", ULONG),
            ("HashOffset", ULONG),
            ("HashFactor", ULONG),
        ]

    class ApiSetHashEntry(ctypes.Structure):
        _fields_ = [
            ("NumberOfEntries", ULONG),
        ]

    class ApiSetHosts_win10(ctypes.Structure):
        _fields_ = [
            ("Hash", ULONG),
            ("Index", ULONG),
        ]

    """
    typedef struct _API_SET_NAMESPACE_ENTRY
    {
        ULONG Flags;
        ULONG NameOffset;
        ULONG NameLength;
        ULONG HashedLength;
        ULONG ValueOffset;
        ULONG ValueCount;
    } API_SET_NAMESPACE_ENTRY, *PAPI_SET_NAMESPACE_ENTRY;
    """

    class ApiSetNameSpaceEntry(ctypes.Structure):
        _fields_ = [
            ("Flags", ULONG),
            ("NameOffset", ULONG),
            ("NameLength", ULONG),
            ("HashedLength", ULONG),
            ("ValueOffset", ULONG),
            ("ValueCount", ULONG),
        ]

    """
    typedef struct _API_SET_VALUE_ENTRY
    {
        ULONG Flags;
        ULONG NameOffset;
        ULONG NameLength;
        ULONG ValueOffset;
        ULONG ValueLength;
    } API_SET_VALUE_ENTRY, *PAPI_SET_VALUE_ENTRY;
    """

    class ApiSetValueEntry(ctypes.Structure):
        _fields_ = [
            ("Flags", ULONG),
            ("NameOffset", ULONG),
            ("NameLength", ULONG),
            ("ValueOffset", ULONG),
            ("ValueLength", ULONG),
        ]


    def get_redirection_by_name(self):
        hdr = self.ApiSetHeader.from_buffer_copy(
            self.api_section[0 : ctypes.sizeof(self.ApiSetHeader)]
        )
        print(repr(hdr))
        redir = {}
        for i in range(hdr.Count):
            addr = hdr.EntryOffset + i * ctypes.sizeof(self.ApiSetNameSpaceEntry)
            entry = self.ApiSetNameSpaceEntry.from_buffer_copy(
                self.api_section[addr : addr + ctypes.sizeof(self.ApiSetNameSpaceEntry)]
            )
            addr = entry.NameOffset
            redir_name = self.api_section[addr : addr + entry.NameLength]
            redir_name = redir_name[::2]
            # print(repr(redir_name))

            addr_descs = entry.ValueOffset

            host_out = {}
            for i in range(entry.ValueCount):
                addr = addr_descs + i * ctypes.sizeof(self.ApiSetValueEntry)
                host = self.ApiSetValueEntry.from_buffer_copy(
                    self.api_section[addr : addr + ctypes.sizeof(self.ApiSetValueEntry)]
                )
                if host.NameOffset != 0:
                    addr = host.NameOffset
                    importName = self.api_section[addr : addr + host.NameLength]
                    importName = importName[::2]
                else:
                    importName = ""
                # print(importName)

                addr = host.ValueOffset
                hostName = self.api_section[addr : addr + host.ValueLength]
                hostName = hostName[::2]
                if not importName:
                    importName = None
                host_out[importName] = hostName
            redir[redir_name.lower()] = host_out

        self.redir = redir

    # hash func can be found in ntdll!ApiSetpSearchForApiSet
    def compute_hash(self, apiset_lib_name, hashf):
        hashk = 0
        for c in apiset_lib_name:
            hashk = (hashk * hashf + c) & ((1 << 32) - 1)
        return hashk

    def get_entry_by_hash(self, dll_name):
        namespace = self.ApiSetHeader.from_buffer_copy(
            self.api_section[: ctypes.sizeof(self.ApiSetHeader)]
        )
        hash_to_search = self.compute_hash(dll_name, namespace.HashFactor)
        print("[+] Looking for hash 0x{0:08X} ({1})".format(hash_to_search, dll_name))
        for i in range(0, namespace.Count):
            offset = namespace.HashOffset + i * ctypes.sizeof(self.ApiSetHosts_win10)
            hash_entry = self.ApiSetHosts_win10.from_buffer_copy(
                self.api_section[
                    offset : offset + ctypes.sizeof(self.ApiSetHosts_win10)
                ]
            )
            if hash_entry.Hash == hash_to_search:
                ##if True:
                print("Hash %X" % hash_entry.Hash)
                offset = namespace.EntryOffset + hash_entry.Index * ctypes.sizeof(
                    self.ApiSetNameSpaceEntry
                )
                namespace_entry = self.ApiSetNameSpaceEntry.from_buffer_copy(
                    self.api_section[
                        offset : offset + ctypes.sizeof(self.ApiSetNameSpaceEntry)
                    ]
                )
                for c in range(0, namespace_entry.ValueCount):
                    print("value_entry[{0}]".format(c))
                    offset = namespace_entry.ValueOffset + c * ctypes.sizeof(
                        self.ApiSetValueEntry
                    )
                    value_entry = self.ApiSetValueEntry.from_buffer_copy(
                        self.api_section[
                            offset : offset + ctypes.sizeof(self.ApiSetValueEntry)
                        ]
                    )
                    value_name = self.api_section[
                        value_entry.NameOffset : value_entry.NameOffset
                        + value_entry.NameLength
                    ]
                    value_data = self.api_section[
                        value_entry.ValueOffset : value_entry.ValueOffset
                        + value_entry.ValueLength
                    ]
                    print("    -> {0}".format(value_name.replace(b"\x00", b"")))
                    print("    -> {0}".format(value_data.replace(b"\x00", b"")))


apiset = ApiSetWin10(sys.argv[1])
apiset.get_redirection_by_name()
dll_name = b"api-ms-win-core-processthreads-l1-1"
apiset.get_entry_by_hash(dll_name)


# fname = sys.argv[1]
# e = PE(open(fname, "rb").read())
# res = e.DirRes.resdesc


# data = extract_409_resource(res)
# token = b"ProductVersion\x00"
# out = []
# for c in token:
#     out += [c, 0]
# out = bytes(out)
# index_start = data.find(out)
# assert index_start > 0
# index_start += len(out)
# index_stop = data.find(b"\x00\x00", index_start)
# assert index_start > 0
# data = data[index_start:index_stop]
# data = data[::2].decode()
# print("Version:", data)

# loc_db = LocationDB()

# apiset.get_entry_by_hash(dll_name)



from builtins import range
from miasm.expression.expression import ExprId
from miasm.core.cpu import reg_info


IP = ExprId('IP', 16)
EIP = ExprId('EIP', 32)
RIP = ExprId('RIP', 64)
exception_flags = ExprId('exception_flags', 32)
interrupt_num = ExprId('interrupt_num', 8)

# GP


regs08_str = ["AL", "CL", "DL", "BL", "AH", "CH", "DH", "BH"] + \
    ["R%dB" % (i + 8) for i in range(8)]
regs08_expr = [ExprId(x, 8) for x in regs08_str]

regs08_64_str = ["AL", "CL", "DL", "BL", "SPL", "BPL", "SIL", "DIL"] + \
    ["R%dB" % (i + 8) for i in range(8)]
regs08_64_expr = [ExprId(x, 8) for x in regs08_64_str]


regs16_str = ["AX", "CX", "DX", "BX", "SP", "BP", "SI", "DI"] + \
    ["R%dW" % (i + 8) for i in range(8)]
regs16_expr = [ExprId(x, 16) for x in regs16_str]

regs32_str = ["EAX", "ECX", "EDX", "EBX", "ESP", "EBP", "ESI", "EDI"] + \
    ["R%dD" % (i + 8) for i in range(8)] + ["EIP"]
regs32_expr = [ExprId(x, 32) for x in regs32_str]

regs64_str = ["RAX", "RCX", "RDX", "RBX", "RSP", "RBP", "RSI", "RDI",
              "R8", "R9", "R10", "R11", "R12", "R13", "R14", "R15",
              "RIP"]
regs64_expr = [ExprId(x, 64) for x in regs64_str]


regs_xmm_str = ["XMM%d" % i for i in range(16)]
regs_xmm_expr = [ExprId(x, 128) for x in regs_xmm_str]

regs_mm_str = ["MM%d" % i for i in range(16)]
regs_mm_expr = [ExprId(x, 64) for x in regs_mm_str]

regs_bnd_str = ["BND%d" % i for i in range(4)]
regs_bnd_expr = [ExprId(x, 128) for x in regs_bnd_str]

gpregs08 = reg_info(regs08_str, regs08_expr)
gpregs08_64 = reg_info(regs08_64_str, regs08_64_expr)
gpregs16 = reg_info(regs16_str, regs16_expr)
gpregs32 = reg_info(regs32_str, regs32_expr)
gpregs64 = reg_info(regs64_str, regs64_expr)

gpregs_xmm = reg_info(regs_xmm_str, regs_xmm_expr)
gpregs_mm = reg_info(regs_mm_str, regs_mm_expr)
gpregs_bnd = reg_info(regs_bnd_str, regs_bnd_expr)

r08_eax = reg_info([regs08_str[0]], [regs08_expr[0]])
r16_eax = reg_info([regs16_str[0]], [regs16_expr[0]])
r32_eax = reg_info([regs32_str[0]], [regs32_expr[0]])
r64_eax = reg_info([regs64_str[0]], [regs64_expr[0]])

r08_ecx = reg_info([regs08_str[1]], [regs08_expr[1]])

r_eax_all = reg_info(
    [regs08_str[0], regs16_str[0], regs32_str[0], regs64_str[0]],
    [regs08_expr[0], regs16_expr[0], regs32_expr[0], regs64_expr[0]])
r_edx_all = reg_info(
    [regs08_str[2], regs16_str[2], regs32_str[2], regs64_str[2]],
    [regs08_expr[2], regs16_expr[2], regs32_expr[2], regs64_expr[2]])

r16_edx = reg_info([regs16_str[2]], [regs16_expr[2]])


selectr_str = ["ES", "CS", "SS", "DS", "FS", "GS"]
selectr_expr = [ExprId(x, 16) for x in selectr_str]
segmreg = reg_info(selectr_str, selectr_expr)

crregs32_str = ["CR%d" % i for i in range(8)]
crregs32_expr = [ExprId(x, 32) for x in crregs32_str]
crregs = reg_info(crregs32_str, crregs32_expr)


drregs32_str = ["DR%d" % i for i in range(8)]
drregs32_expr = [ExprId(x, 32) for x in drregs32_str]
drregs = reg_info(drregs32_str, drregs32_expr)


fltregs32_str = ["ST(%d)" % i for i in range(8)]
fltregs32_expr = [ExprId(x, 64) for x in fltregs32_str]
fltregs = reg_info(fltregs32_str, fltregs32_expr)

r_st_all = reg_info(['ST'],
                    [ExprId('ST', 64)])

r_cs_all = reg_info(['CS'],
                    [ExprId('CS', 16)])
r_ds_all = reg_info(['DS'],
                    [ExprId('DS', 16)])
r_es_all = reg_info(['ES'],
                    [ExprId('ES', 16)])
r_ss_all = reg_info(['SS'],
                    [ExprId('SS', 16)])
r_fs_all = reg_info(['FS'],
                    [ExprId('FS', 16)])
r_gs_all = reg_info(['GS'],
                    [ExprId('GS', 16)])


AL = regs08_expr[0]
CL = regs08_expr[1]
DL = regs08_expr[2]
BL = regs08_expr[3]
AH = regs08_expr[4]
CH = regs08_expr[5]
DH = regs08_expr[6]
BH = regs08_expr[7]
R8B = regs08_expr[8]
R9B = regs08_expr[9]
R10B = regs08_expr[10]
R11B = regs08_expr[11]
R12B = regs08_expr[12]
R13B = regs08_expr[13]
R14B = regs08_expr[14]
R15B = regs08_expr[15]

SPL = regs08_64_expr[4]
BPL = regs08_64_expr[5]
SIL = regs08_64_expr[6]
DIL = regs08_64_expr[7]


AX = regs16_expr[0]
CX = regs16_expr[1]
DX = regs16_expr[2]
BX = regs16_expr[3]
SP = regs16_expr[4]
BP = regs16_expr[5]
SI = regs16_expr[6]
DI = regs16_expr[7]
R8W = regs16_expr[8]
R9W = regs16_expr[9]
R10W = regs16_expr[10]
R11W = regs16_expr[11]
R12W = regs16_expr[12]
R13W = regs16_expr[13]
R14W = regs16_expr[14]
R15W = regs16_expr[15]


EAX = regs32_expr[0]
ECX = regs32_expr[1]
EDX = regs32_expr[2]
EBX = regs32_expr[3]
ESP = regs32_expr[4]
EBP = regs32_expr[5]
ESI = regs32_expr[6]
EDI = regs32_expr[7]
R8D = regs32_expr[8]
R9D = regs32_expr[9]
R10D = regs32_expr[10]
R11D = regs32_expr[11]
R12D = regs32_expr[12]
R13D = regs32_expr[13]
R14D = regs32_expr[14]
R15D = regs32_expr[15]


RAX = regs64_expr[0]
RCX = regs64_expr[1]
RDX = regs64_expr[2]
RBX = regs64_expr[3]
RSP = regs64_expr[4]
RBP = regs64_expr[5]
RSI = regs64_expr[6]
RDI = regs64_expr[7]
R8 = regs64_expr[8]
R9 = regs64_expr[9]
R10 = regs64_expr[10]
R11 = regs64_expr[11]
R12 = regs64_expr[12]
R13 = regs64_expr[13]
R14 = regs64_expr[14]
R15 = regs64_expr[15]


reg_zf = 'zf'
reg_nf = 'nf'
reg_pf = 'pf'
reg_of = 'of'
reg_cf = 'cf'
reg_tf = 'tf'
reg_if = 'i_f'
reg_df = 'df'
reg_af = 'af'
reg_iopl = 'iopl_f'
reg_nt = 'nt'
reg_rf = 'rf'
reg_vm = 'vm'
reg_ac = 'ac'
reg_vif = 'vif'
reg_vip = 'vip'
reg_id = 'i_d'


reg_es = "ES"
reg_cs = "CS"
reg_ss = "SS"
reg_ds = "DS"
reg_fs = "FS"
reg_gs = "GS"

reg_dr0 = 'DR0'
reg_dr1 = 'DR1'
reg_dr2 = 'DR2'
reg_dr3 = 'DR3'
reg_dr4 = 'DR4'
reg_dr5 = 'DR5'
reg_dr6 = 'DR6'
reg_dr7 = 'DR7'

reg_cr0 = 'CR0'
reg_cr1 = 'CR1'
reg_cr2 = 'CR2'
reg_cr3 = 'CR3'
reg_cr4 = 'CR4'
reg_cr5 = 'CR5'
reg_cr6 = 'CR6'
reg_cr7 = 'CR7'

reg_mm0 = 'MM0'
reg_mm1 = 'MM1'
reg_mm2 = 'MM2'
reg_mm3 = 'MM3'
reg_mm4 = 'MM4'
reg_mm5 = 'MM5'
reg_mm6 = 'MM6'
reg_mm7 = 'MM7'

reg_tsc = "tsc"

reg_float_c0 = 'float_c0'
reg_float_c1 = 'float_c1'
reg_float_c2 = 'float_c2'
reg_float_c3 = 'float_c3'
reg_float_stack_ptr = "float_stack_ptr"
reg_float_control = 'reg_float_control'
reg_float_eip = 'reg_float_eip'
reg_float_cs = 'reg_float_cs'
reg_float_address = 'reg_float_address'
reg_float_ds = 'reg_float_ds'


dr0 = ExprId(reg_dr0, 32)
dr1 = ExprId(reg_dr1, 32)
dr2 = ExprId(reg_dr2, 32)
dr3 = ExprId(reg_dr3, 32)
dr4 = ExprId(reg_dr4, 32)
dr5 = ExprId(reg_dr5, 32)
dr6 = ExprId(reg_dr6, 32)
dr7 = ExprId(reg_dr7, 32)

cr0 = ExprId(reg_cr0, 32)
cr1 = ExprId(reg_cr1, 32)
cr2 = ExprId(reg_cr2, 32)
cr3 = ExprId(reg_cr3, 32)
cr4 = ExprId(reg_cr4, 32)
cr5 = ExprId(reg_cr5, 32)
cr6 = ExprId(reg_cr6, 32)
cr7 = ExprId(reg_cr7, 32)

mm0 = ExprId(reg_mm0, 64)
mm1 = ExprId(reg_mm1, 64)
mm2 = ExprId(reg_mm2, 64)
mm3 = ExprId(reg_mm3, 64)
mm4 = ExprId(reg_mm4, 64)
mm5 = ExprId(reg_mm5, 64)
mm6 = ExprId(reg_mm6, 64)
mm7 = ExprId(reg_mm7, 64)

XMM0 = regs_xmm_expr[0]
XMM1 = regs_xmm_expr[1]
XMM2 = regs_xmm_expr[2]
XMM3 = regs_xmm_expr[3]
XMM4 = regs_xmm_expr[4]
XMM5 = regs_xmm_expr[5]
XMM6 = regs_xmm_expr[6]
XMM7 = regs_xmm_expr[7]
XMM8 = regs_xmm_expr[8]
XMM9 = regs_xmm_expr[9]
XMM10 = regs_xmm_expr[10]
XMM11 = regs_xmm_expr[11]
XMM12 = regs_xmm_expr[12]
XMM13 = regs_xmm_expr[13]
XMM14 = regs_xmm_expr[14]
XMM15 = regs_xmm_expr[15]

# tmp1= ExprId(reg_tmp1)
zf = ExprId(reg_zf, size=1)
nf = ExprId(reg_nf, size=1)
pf = ExprId(reg_pf, size=1)
of = ExprId(reg_of, size=1)
cf = ExprId(reg_cf, size=1)
tf = ExprId(reg_tf, size=1)
i_f = ExprId(reg_if, size=1)
df = ExprId(reg_df, size=1)
af = ExprId(reg_af, size=1)
iopl = ExprId(reg_iopl, size=2)
nt = ExprId(reg_nt, size=1)
rf = ExprId(reg_rf, size=1)
vm = ExprId(reg_vm, size=1)
ac = ExprId(reg_ac, size=1)
vif = ExprId(reg_vif, size=1)
vip = ExprId(reg_vip, size=1)
i_d = ExprId(reg_id, size=1)

ES = ExprId(reg_es, size=16)
CS = ExprId(reg_cs, size=16)
SS = ExprId(reg_ss, size=16)
DS = ExprId(reg_ds, size=16)
FS = ExprId(reg_fs, size=16)
GS = ExprId(reg_gs, size=16)

tsc = ExprId(reg_tsc, size=64)

float_c0 = ExprId(reg_float_c0, size=1)
float_c1 = ExprId(reg_float_c1, size=1)
float_c2 = ExprId(reg_float_c2, size=1)
float_c3 = ExprId(reg_float_c3, size=1)
float_stack_ptr = ExprId(reg_float_stack_ptr, size=3)
float_control = ExprId(reg_float_control, 16)
float_eip = ExprId(reg_float_eip, 32)
float_cs = ExprId(reg_float_cs, size=16)
float_address = ExprId(reg_float_address, 32)
float_ds = ExprId(reg_float_ds, size=16)

float_st0 = ExprId("float_st0", 64)
float_st1 = ExprId("float_st1", 64)
float_st2 = ExprId("float_st2", 64)
float_st3 = ExprId("float_st3", 64)
float_st4 = ExprId("float_st4", 64)
float_st5 = ExprId("float_st5", 64)
float_st6 = ExprId("float_st6", 64)
float_st7 = ExprId("float_st7", 64)


float_list = [float_st0, float_st1, float_st2, float_st3,
              float_st4, float_st5, float_st6, float_st7]

float_replace = {fltregs32_expr[i]: float_list[i] for i in range(8)}
float_replace[r_st_all.expr[0]] = float_st0


EAX_init = ExprId('EAX_init', 32)
EBX_init = ExprId('EBX_init', 32)
ECX_init = ExprId('ECX_init', 32)
EDX_init = ExprId('EDX_init', 32)
ESI_init = ExprId('ESI_init', 32)
EDI_init = ExprId('EDI_init', 32)
ESP_init = ExprId('ESP_init', 32)
EBP_init = ExprId('EBP_init', 32)


RAX_init = ExprId('RAX_init', 64)
RBX_init = ExprId('RBX_init', 64)
RCX_init = ExprId('RCX_init', 64)
RDX_init = ExprId('RDX_init', 64)
RSI_init = ExprId('RSI_init', 64)
RDI_init = ExprId('RDI_init', 64)
RSP_init = ExprId('RSP_init', 64)
RBP_init = ExprId('RBP_init', 64)


all_regs_ids = [
    AL, CL, DL, BL, AH, CH, DH, BH,
    R8B, R9B, R10B, R11B, R12B, R13B, R14B, R15B,
    SPL, BPL, SIL, DIL,
    AX, CX, DX, BX, SP, BP, SI, DI,
    R8W, R9W, R10W, R11W, R12W, R13W, R14W, R15W,
    IP,
    EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI,
    R8D, R9D, R10D, R11D, R12D, R13D, R14D, R15D,
    EIP,

    RAX, RBX, RCX, RDX, RSP, RBP, RIP, RSI, RDI,
    R8, R9, R10, R11, R12, R13, R14, R15,
    zf, nf, pf, of, cf, af, df,
    tf, i_f, iopl, nt, rf, vm, ac, vif, vip, i_d,
    float_control, float_eip, float_cs, float_address, float_ds,
    tsc,
    ES, CS, SS, DS, FS, GS,
    float_st0, float_st1, float_st2, float_st3,
    float_st4, float_st5, float_st6, float_st7,
    float_c0, float_c1, float_c2, float_c3,
    cr0, cr3,
    dr0, dr1, dr2, dr3, dr4, dr5, dr6, dr7,
    float_stack_ptr,
    mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7,

    XMM0, XMM1, XMM2, XMM3, XMM4, XMM5, XMM6, XMM7,
    XMM8, XMM9, XMM10, XMM11, XMM12, XMM13, XMM14, XMM15,


    exception_flags, interrupt_num,
] + fltregs32_expr

all_regs_ids_no_alias = [
    RAX, RBX, RCX, RDX, RSP, RBP, RIP, RSI, RDI,
    R8, R9, R10, R11, R12, R13, R14, R15,
    zf, nf, pf, of, cf, af, df,
    tf, i_f, iopl, nt, rf, vm, ac, vif, vip, i_d,
    float_control, float_eip, float_cs, float_address, float_ds,
    tsc,
    ES, CS, SS, DS, FS, GS,
    float_st0, float_st1, float_st2, float_st3,
    float_st4, float_st5, float_st6, float_st7,
    float_c0, float_c1, float_c2, float_c3,
    cr0, cr3,
    dr0, dr1, dr2, dr3, dr4, dr5, dr6, dr7,
    float_stack_ptr,
    mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7,
    XMM0, XMM1, XMM2, XMM3, XMM4, XMM5, XMM6, XMM7,
    XMM8, XMM9, XMM10, XMM11, XMM12, XMM13, XMM14, XMM15,


    exception_flags, interrupt_num,
] + fltregs32_expr

attrib_to_regs = {
    16: regs16_expr + all_regs_ids_no_alias[all_regs_ids_no_alias.index(zf):] + [IP],
    32: regs32_expr + all_regs_ids_no_alias[all_regs_ids_no_alias.index(zf):] + [EIP],
    64: all_regs_ids_no_alias,
}

all_regs_ids_byname = dict([(x.name, x) for x in all_regs_ids])

all_regs_ids_init = [ExprId("%s_init" % x.name, x.size) for x in all_regs_ids]

regs_init = {}
for i, r in enumerate(all_regs_ids):
    regs_init[r] = all_regs_ids_init[i]

regs_flt_expr = [float_st0, float_st1, float_st2, float_st3,
                 float_st4, float_st5, float_st6, float_st7,
                 ]

mRAX = {16: AX, 32: EAX, 64: RAX}
mRBX = {16: BX, 32: EBX, 64: RBX}
mRCX = {16: CX, 32: ECX, 64: RCX}
mRDX = {16: DX, 32: EDX, 64: RDX}
mRSI = {16: SI, 32: ESI, 64: RSI}
mRDI = {16: DI, 32: EDI, 64: RDI}
mRBP = {16: BP, 32: EBP, 64: RBP}
mRSP = {16: SP, 32: ESP, 64: RSP}
mRIP = {16: IP, 32: EIP, 64: RIP}



from miasm.expression.expression import ExprAssign, ExprOp
from miasm.ir.ir import AssignBlock
from miasm.ir.analysis import LifterModelCall
from miasm.arch.ppc.sem import Lifter_PPC32b


class LifterModelCallPpc32b(Lifter_PPC32b, LifterModelCall):

    def __init__(self, loc_db, *args):
        super(LifterModelCallPpc32b, self).__init__(loc_db, *args)
        self.ret_reg = self.arch.regs.R3

    # for test XXX TODO
    def set_dead_regs(self, irblock):
        pass

    def get_out_regs(self, _):
        return set([self.ret_reg, self.sp])

    def add_unused_regs(self):
        leaves = [self.blocks[label] for label in self.g.leafs()]
        for irblock in leaves:
            self.set_dead_regs(irblock)

    def call_effects(self, ad, instr):
        call_assignblks = AssignBlock(
            [
                ExprAssign(
                    self.ret_reg,
                    ExprOp(
                        'call_func_ret',
                        ad,
                        self.sp,
                        self.arch.regs.R3,
                        self.arch.regs.R4,
                        self.arch.regs.R5,
                    )
                ),
                ExprAssign(self.sp, ExprOp('call_func_stack', ad, self.sp)),
            ],
            instr
        )
        return [call_assignblks], []

    def add_instr_to_current_state(self, instr, block, assignments, ir_blocks_all, gen_pc_updt):
        """
        Add the IR effects of an instruction to the current state.

        @instr: native instruction
        @block: native block source
        @assignments: list of current AssignBlocks
        @ir_blocks_all: list of additional effects
        @gen_pc_updt: insert PC update effects between instructions
        """
        if instr.is_subcall():
            call_assignblks, extra_irblocks = self.call_effects(
                instr.getdstflow(None)[0],
                instr
            )
            assignments += call_assignblks
            ir_blocks_all += extra_irblocks
            return True

        if gen_pc_updt is not False:
            self.gen_pc_update(assignments, instr)

        assignblk, ir_blocks_extra = self.instr2ir(instr)
        assignments.append(assignblk)
        ir_blocks_all += ir_blocks_extra
        if ir_blocks_extra:
            return True
        return False

    def sizeof_char(self):
        return 8

    def sizeof_short(self):
        return 16

    def sizeof_int(self):
        return 32

    def sizeof_long(self):
        return 32

    def sizeof_pointer(self):
        return 32



from argparse import ArgumentParser
from miasm.jitter.csts import PAGE_READ, PAGE_WRITE
from miasm.analysis.machine import Machine
from miasm.core.locationdb import LocationDB

from pdb import pm

parser = ArgumentParser(description="x86 32 basic Jitter")
parser.add_argument("filename", help="x86 32 shellcode filename")
parser.add_argument("-j", "--jitter",
                    help="Jitter engine (default is 'gcc')",
                    default="gcc")
args = parser.parse_args()

def code_sentinelle(jitter):
    jitter.running = False
    jitter.pc = 0
    return True

loc_db = LocationDB()

myjit = Machine("x86_32").jitter(loc_db, args.jitter)
myjit.init_stack()

data = open(args.filename, 'rb').read()
run_addr = 0x40000000
myjit.vm.add_memory_page(run_addr, PAGE_READ | PAGE_WRITE, data)

myjit.set_trace_log()
myjit.push_uint32_t(0x1337beef)

myjit.add_breakpoint(0x1337beef, code_sentinelle)

myjit.run(run_addr)



from __future__ import print_function
from miasm.expression.expression import *
from miasm.expression.simplifications import expr_simp

print("""
Simple expression simplification demo
""")


a = ExprId('eax', 32)
b = ExprId('ebx', 32)

exprs = [a + b - a,
         ExprInt(0x12, 32) + ExprInt(0x30, 32) - a,
         ExprCompose(a[:8], a[8:16])]

for e in exprs:
    print('*' * 40)
    print('original expression:', e)
    print("simplified:", expr_simp(e))



from __future__ import print_function
from builtins import zip
import miasm.jitter.jitcore as jitcore
from miasm.expression.expression import ExprInt, ExprLoc
import miasm.jitter.csts as csts
from miasm.expression.simplifications import expr_simp_explicit
from miasm.jitter.emulatedsymbexec import EmulatedSymbExec

################################################################################
#                              Python jitter Core                              #
################################################################################


class JitCore_Python(jitcore.JitCore):
    "JiT management, using Miasm2 Symbol Execution engine as backend"

    SymbExecClass = EmulatedSymbExec

    def __init__(self, lifter, bin_stream):
        super(JitCore_Python, self).__init__(lifter, bin_stream)
        self.lifter = lifter
        self.ircfg = self.lifter.new_ircfg()

        # CPU & VM (None for now) will be set later

        self.symbexec = self.SymbExecClass(
            None, None,
            self.lifter, {},
            sb_expr_simp=expr_simp_explicit
        )
        self.symbexec.enable_emulated_simplifications()

    def set_cpu_vm(self, cpu, vm):
        self.symbexec.cpu = cpu
        self.symbexec.vm = vm

    def load(self):
        "Preload symbols according to current architecture"
        self.symbexec.reset_regs()

    def arch_specific(self):
        """Return arch specific information for the current architecture"""
        arch = self.lifter.arch
        has_delayslot = False
        if arch.name == "mips32":
            from miasm.arch.mips32.jit import mipsCGen
            cgen_class = mipsCGen
            has_delayslot = True
        elif arch.name == "arm":
            from miasm.arch.arm.jit import arm_CGen
            cgen_class = arm_CGen
        else:
            from miasm.jitter.codegen import CGen
            cgen_class = CGen
        return cgen_class(self.lifter), has_delayslot

    def add_block(self, asmblock):
        """Create a python function corresponding to an AsmBlock
        @asmblock: AsmBlock
        """

        # TODO: merge duplicate code with CGen, llvmconvert
        codegen, has_delayslot = self.arch_specific()
        irblocks_list = codegen.block2assignblks(asmblock)
        instr_offsets = [line.offset for line in asmblock.lines]

        loc_db = self.lifter.loc_db
        local_loc_keys = []
        for irblocks in irblocks_list:
            for irblock in irblocks:
                local_loc_keys.append(irblock.loc_key)

        def myfunc(cpu):
            """Execute the function according to cpu and vmmngr states
            @cpu: JitCpu instance
            """
            # Get virtual memory handler
            vmmngr = cpu.vmmngr

            # Get execution engine (EmulatedSymbExec instance)
            exec_engine = self.symbexec

            # Refresh CPU values according to @cpu instance
            exec_engine.update_engine_from_cpu()

            # Get initial loc_key
            cur_loc_key = asmblock.loc_key

            # Update PC helper
            update_pc = lambda value: setattr(cpu, self.lifter.pc.name, value)

            while True:
                # Retrieve the expected irblock
                for instr, irblocks in zip(asmblock.lines, irblocks_list):
                    for index, irblock in enumerate(irblocks):
                        if irblock.loc_key == cur_loc_key:
                            break
                    else:
                        continue
                    break
                else:
                    raise RuntimeError("Unable to find the block for %r" % cur_loc_key)

                instr_attrib, irblocks_attributes = codegen.get_attributes(
                    instr, irblocks, self.log_mn, self.log_regs
                )
                irblock_attributes = irblocks_attributes[index]

                # Do IRBlock
                new_irblock = self.lifter.irbloc_fix_regs_for_mode(
                    irblock, self.lifter.attrib
                )
                if index == 0:
                    # Pre code
                    if instr_attrib.log_mn:
                        print("%.8X %s" % (
                            instr_attrib.instr.offset,
                            instr_attrib.instr.to_string(loc_db)
                        ))

                # Exec IRBlock
                instr = instr_attrib.instr

                for index, assignblk in enumerate(irblock):
                    attributes = irblock_attributes[index]

                    # Eval current instruction (in IR)
                    exec_engine.eval_updt_assignblk(assignblk)

                    # Check memory access / write exception
                    # TODO: insert a check between memory reads and writes
                    if attributes.mem_read or attributes.mem_write:
                        # Restricted exception
                        flag = ~csts.EXCEPT_CODE_AUTOMOD & csts.EXCEPT_DO_NOT_UPDATE_PC
                        if (vmmngr.get_exception() & flag != 0):
                            # Do not update registers
                            update_pc(instr.offset)
                            return instr.offset

                    # Update registers values
                    exec_engine.update_cpu_from_engine()

                    # Check post assignblk exception flags
                    if attributes.set_exception:
                        # Restricted exception
                        if cpu.get_exception() > csts.EXCEPT_NUM_UPDT_EIP:
                            # Update PC
                            update_pc(instr.offset)
                            return instr.offset

                dst = exec_engine.eval_expr(self.lifter.IRDst)
                if dst.is_int():
                    loc_key = loc_db.get_or_create_offset_location(int(dst))
                    dst = ExprLoc(loc_key, dst.size)

                assert dst.is_loc()
                loc_key = dst.loc_key
                offset = loc_db.get_location_offset(loc_key)
                if offset is None:
                    # Avoid checks on generated label
                    cur_loc_key = loc_key
                    continue

                if instr_attrib.log_regs:
                    update_pc(offset)
                    cpu.dump_gpregs_with_attrib(self.lifter.attrib)

                # Post-instr checks
                if instr_attrib.mem_read | instr_attrib.mem_write:
                    vmmngr.check_memory_breakpoint()
                    vmmngr.check_invalid_code_blocs()
                    if vmmngr.get_exception():
                        update_pc(offset)
                        return offset

                if instr_attrib.set_exception:
                    if cpu.get_exception():
                        update_pc(offset)
                        return offset

                if instr_attrib.mem_read | instr_attrib.mem_write:
                    vmmngr.reset_memory_access()

                # Manage resulting address
                if (loc_key in local_loc_keys and
                    offset > instr.offset):
                    # Forward local jump
                    # Note: a backward local jump has to be promoted to extern,
                    # for max_exec_per_call support
                    cur_loc_key = loc_key
                    continue

                # Delay slot
                if has_delayslot:
                    delay_slot_set = exec_engine.eval_expr(codegen.delay_slot_set)
                    if delay_slot_set.is_int() and int(delay_slot_set) != 0:
                        return int(exec_engine.eval_expr(codegen.delay_slot_dst))

                # Extern of asmblock, must have an offset
                assert offset is not None
                return offset

        # Associate myfunc with current loc_key
        offset = loc_db.get_location_offset(asmblock.loc_key)
        assert offset is not None
        self.offset_to_jitted_func[offset] = myfunc

    def exec_wrapper(self, loc_key, cpu, _offset_to_jitted_func, _stop_offsets,
                     _max_exec_per_call):
        """Call the function @loc_key with @cpu
        @loc_key: function's loc_key
        @cpu: JitCpu instance
        """

        # Get Python function corresponding to @loc_key
        fc_ptr = self.offset_to_jitted_func[loc_key]

        # Execute the function
        return fc_ptr(cpu)
